import unittest
import yaz

from test.extension.dependencyinjection import Food, MoreFood
from test.extension.pluginordinal import base as Orginal


class TestPlugin(unittest.TestCase):
    def test_010_dependency_injection(self):
        """Should call dependency setter"""
        food = yaz.get_plugin_instance(Food)
        self.assertEqual("Breakfast is ready", food.breakfast())
        self.assertEqual("BREAKFAST IS READY", food.breakfast(shout=True))

    def test_020_singleton_dependency_injection(self):
        """Should inject the same dependency for different dependent plugins"""
        food = yaz.get_plugin_instance(Food)
        more_food = yaz.get_plugin_instance(MoreFood)
        self.assertEqual(food.get_helper(), more_food.get_helper())

    def test_030_plugin_creation_ordinal(self):
        """Should follow the ordinal when creating the plugin type"""
        ordinal = yaz.get_plugin_instance(Orginal)
        self.assertEqual(["low (8)", "custom (128)", "default (256)", "base (512)", "high (1024)"], ordinal.stack())



def decorator(decorator_func):
    """Allows a decorator to be called with or without keyword arguments."""
    assert callable(decorator_func), type(decorator_func)

    def _decorator(func=None, **kwargs):
        assert func is None or callable(func), type(func)
        if func:
            return decorator_func(func, **kwargs)
        else:
            def _decorator_helper(func):
                return decorator_func(func, **kwargs)

            return _decorator_helper

    return _decorator

# @decorator
# def cache(func, ignore=None, type='memory', sqlite_filename='~/.yaz-cache.sqlite'):
#     assert ignore is None or isinstance(ignore, (tuple, list)), type(ignore)

#     def generate_key(args, kwargs):
#         # todo: do not cache when args and kwargs are not collections.Hashable (i.e. when we can not
#         # generate a proper key)
#         return (args, tuple(sorted((key, value) for key, value in kwargs.items() if not key in ignore)))

#     @functools.wraps(func)
#     def cache_from_memory(self, *args, **kwargs):
#         key = generate_key(args, kwargs)
#         hit = key in cache
#         if self.yaz.verbose:
#             print(self.render("{% color '', '', 'reverse' %}--- {{ plugin.__class__.__name__ }} {{ task.__name__ }} {{ cache }}{% endcolor %} {{ sourcefile }}",
#                               dict(plugin=self, task=func, sourcefile=inspect.getsourcefile(func), cache="CACHE HIT" if hit else "CACHE MISS")))
#         if not hit:
#             cache[key] = func(self, *args, **kwargs)
#         return cache[key]

#     @functools.wraps(func)
#     def cache_from_sqlite(self, *args, **kwargs):
#         key = generate_key(args, kwargs)
#         hit = key in cache
#         if self.yaz.verbose:
#             print(self.render("{% color '', '', 'reverse' %}--- {{ plugin.__class__.__name__ }} {{ task.__name__ }} {{ cache }}{% endcolor %} {{ sourcefile }}",
#                               dict(plugin=self, task=func, sourcefile=inspect.getsourcefile(func), cache="CACHE HIT" if hit else "CACHE MISS")))
#         if not hit:
#             cache[key] = func(self, *args, **kwargs)
#         return cache[key]

#     if type == 'memory':
#         cache = dict()
#         return cache_from_memory

#     if type == 'sqlite':
#         return cache_from_sqlite

#     raise RuntimeException('The parameter type must be either "memory" or "sqlite"')



import sys

from .loader import load
from .parser import Parser
from .task import get_task_tree


def main(argv=None, white_list=None, load_yaz_extension=True):
    assert argv is None or isinstance(argv, list), type(argv)
    assert white_list is None or isinstance(white_list, list), type(white_list)
    assert isinstance(load_yaz_extension, bool), type(load_yaz_extension)

    if load_yaz_extension:
        load("~/.yaz", "yaz_extension")

    parser = Parser()
    parser.add_task_tree(get_task_tree(white_list))

    task, kwargs = parser.parse_arguments(sys.argv if argv is None else argv)
    if task:
        result = task(**kwargs)
        if result is not None:
            print(result)

        if isinstance(result, int) and not isinstance(result, bool):
            exit(result)

        return result

    else:
        parser.print_help()



import asyncio
import yaz

import test.extension.coroutine as co_routine


class TestTask(yaz.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.caller = self.get_caller([co_routine.Coroutine])
        self.loop = asyncio.get_event_loop()
        self.sleep = 0.1
        self.delta = self.sleep * 0.25

    def test_010_one(self):
        """Should call one task asynchronously"""
        start = self.loop.time()
        self.assertEqual(self.sleep, self.caller("do-one", str(self.sleep)))
        self.assertAlmostEqual(self.sleep, self.loop.time() - start, delta=self.delta)

    def test_020_many(self):
        """Should call multiple tasks asynchronously"""
        start = self.loop.time()
        self.assertEqual([self.sleep for _ in range(10)], self.caller("do-many", "10", str(self.sleep)))
        self.assertAlmostEqual(self.sleep, self.loop.time() - start, delta=self.delta)



import os.path
import yaz

import test.extension.singlefunction as single_function
import test.extension.multiplefunctions as multiple_functions
import test.extension.singlemethod as single_method
import test.extension.multiplemethods as multiple_methods
import test.extension.namingconvention as naming_convention
import test.extension.typeannotation as type_annotation
import test.extension.taskconfiguration as task_configuration


class TestParser(yaz.TestCase):
    def test_010_single_function(self):
        """Should be available without specifying the task name"""
        caller = self.get_caller([single_function.say])

        # without arguments, uses default value for message
        self.assertEqual("Hello World!", caller())

        # provide message value as an argument
        self.assertEqual("I said HELLO!", caller("--message", "I said HELLO!"))

    def test_020_multiple_functions(self):
        """Should be available after specifying the task name"""
        caller = self.get_caller([multiple_functions.foo, multiple_functions.bar])

        # without arguments, can not determine the task to run
        with self.assertRaisesRegex(AssertionError, "The parser did not find a task"):
            caller()

        # provide the task as an argument
        self.assertEqual("Foo", caller("foo"))

        # provide the task as an argument
        self.assertEqual("Bar", caller("bar"))

    def test_030_single_method(self):
        """Should be available without specifying the plugin nor the task name"""
        caller = self.get_caller([single_method.Person])
        self.assertEqual("I have very little to say.", caller())

    def test_040_multiple_methods(self):
        """Should be available without specifying the plugin name but after specifying the task name"""
        caller = self.get_caller([multiple_methods.Shape])

        # without arguments, can not determine the task to run
        with self.assertRaisesRegex(AssertionError, "The parser did not find a task"):
            caller()

        # provide the task as an argument
        self.assertEqual("Circle", caller("circle"))

        # provide the task as an argument
        self.assertEqual("Square", caller("square"))

    def test_050_multiple_plugins(self):
        """Should be available after specifying the plugin name and task name"""
        caller = self.get_caller([multiple_methods.Shape, single_method.Person])

        # without arguments, can not determine the task to run
        with self.assertRaisesRegex(AssertionError, "The parser did not find a task"):
            caller()

        # provide the plugin as an argument (task name is not needed, as Person only has one task)
        self.assertEqual("I have very little to say.", caller("person"))

        # provide the plugin and task as an argument
        self.assertEqual("Circle", caller("shape", "circle"))

        # provide the plugin and task as an argument
        self.assertEqual("Square", caller("shape", "square"))

    def test_060_naming_convention(self):
        """Should convert Plugin name, task name, and argument name to conform with naming conventions"""
        caller = self.get_caller([naming_convention.This_was__a_weirdMIXTure__, naming_convention.ThisWasCamelCase])
        self.assertEqual("this-was-a-weird-mixture", caller("this-was-a-weird-mixture"))
        self.assertEqual("this-was-camel-case", caller("this-was-camel-case", "expected-class-name"))
        self.assertEqual("this-was-underscored", caller("this-was-camel-case", "this-was-underscored"))
        self.assertEqual("this-was-camel-case", caller("this-was-camel-case", "this-was-camel-case"))
        self.assertEqual("this-was-also-underscored", caller("this-was-camel-case", "this-was-also-underscored"))
        self.assertEqual(("a", "b", "c", "d"), caller("this-was-camel-case", "required-arguments", "a", "b", "c", "d"))
        self.assertEqual(("A", "B", "C", "D"), caller("this-was-camel-case", "optional-arguments"))

    def test_070_string_type_annotation(self):
        """Should understand string type annotation"""
        caller = self.get_caller([type_annotation.TypeAnnotation])
        self.assertEqual("Hello World", caller("required-string", "Hello World"))
        self.assertEqual("Alternative", caller("optional-string", "--string=Alternative"))
        self.assertEqual("Hello World!", caller("optional-string"))

    def test_080_boolean_type_annotation(self):
        """Should understand boolean type annotation"""
        caller = self.get_caller([type_annotation.TypeAnnotation])
        self.assertTrue(caller("required-boolean", "--check"))
        self.assertFalse(caller("required-boolean", "--no-check"))
        self.assertTrue(caller("optional-boolean-true"))
        self.assertFalse(caller("optional-boolean-true", "--no-check"))
        self.assertFalse(caller("optional-boolean-false"))
        self.assertTrue(caller("optional-boolean-false", "--check"))

    def test_090_integer_type_annotation(self):
        """Should understand integer type annotation"""
        caller = self.get_caller([type_annotation.TypeAnnotation])
        self.assertEqual(123, caller("required-integer", "123"))
        self.assertEqual(42, caller("optional-integer"))
        self.assertEqual(123, caller("optional-integer", "--number", "123"))

    def test_100_float_type_annotation(self):
        """Should understand float type annotation"""
        caller = self.get_caller([type_annotation.TypeAnnotation])
        self.assertEqual(0.5, caller("required-float", "0.5"))
        self.assertEqual(3.14, caller("optional-float"))
        self.assertEqual(0.5, caller("optional-float", "--number", "0.5"))

    def test_110_file_type_annotation(self):
        """Should understand file, i.e. open, type annotation"""
        caller = self.get_caller([type_annotation.TypeAnnotation])

        expected = "First line\nSecond line\nThird line"
        self.assertEqual(expected, caller("required-file", os.path.join(os.path.dirname(__file__), "file_type_annotation/input.txt")))
        self.assertEqual(expected, caller("optional-file", "--file", os.path.join(os.path.dirname(__file__), "file_type_annotation/input.txt")))

        expected = "#!/usr/bin/env python3"
        self.assertEqual(
            expected,
            caller("optional-file", "--length", str(len(expected))))

    def test_120_choices_configuration(self):
        """Should accept predefined choices"""
        caller = self.get_caller([task_configuration.ConfigurationPlugin])

        # using available choice
        self.assertTrue(caller("required-choice", "yes"))

        # using unavailable choice
        with self.assertRaises(SystemExit):
            caller("required-choice", "unavailable")



import sys
import unittest

from yaz import main


class TestMain(unittest.TestCase):
    def test_main(self):
        """Should load a module from a given directory"""
        result = main([sys.argv[0], "--message", "Echo!"], ["say"])
        self.assertEqual("Echo!", result)

    def test_main_runs_help(self):
        """Should print help when not given arguments"""
        # todo: check stdout
        main([sys.argv[0]])


import re
import inspect
import asyncio

from .decorator import decorator
from .plugin import BasePlugin, get_plugin_instance, get_plugin_list


class Task:
    class Documentation:
        def __init__(self, full):
            if full:
                match = re.match(r"^(?P<first_line>.+)\n\n(?P<everything_else>.+)$", full, re.MULTILINE)
                if match:
                    short = match.group("first_line")
                    long = match.group("everything_else")
                else:
                    short = full
                    long = ""
            else:
                full = ""
                short = ""
                long = ""

            self.full = full.strip()
            self.short = short.strip()
            self.long = long.strip()

    def __init__(self, plugin_class, func, config):
        assert plugin_class is None or issubclass(plugin_class, BasePlugin)
        assert callable(func)
        assert isinstance(config, dict)
        self.plugin_class = plugin_class
        self.func = func
        self.config = config
        self.documentation = self.Documentation(inspect.getdoc(self.func))

    def __call__(self, **kwargs):
        """Prepare dependencies and call this Task"""
        if self.plugin_class:
            result = self.func(get_plugin_instance(self.plugin_class), **kwargs)

        else:
            result = self.func(**kwargs)

        if inspect.iscoroutinefunction(self.func):
            assert inspect.iscoroutine(
                result), "The task is defined as a coroutine function but does not return a coroutine"
            loop = asyncio.get_event_loop()
            if loop.is_closed():
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
            result = loop.run_until_complete(result)
            loop.close()

        return result

    def get_qualified_name(self):
        """Returns the __qualname__ of this Task"""
        return self.func.__qualname__

    def get_parameters(self):
        """Returns a list of parameters"""
        sig = inspect.signature(self.func)
        for index, parameter in enumerate(sig.parameters.values()):
            if index == 0 and self.plugin_class is not None:
                # skip SELF parameter
                continue

            yield parameter

    def get_configuration(self, key, default=None):
        """Returns the configuration for KEY"""
        if key in self.config:
            return self.config.get(key)
        else:
            return default

    def __str__(self):
        return self.get_qualified_name()


_task_list = {}


def get_task_tree(white_list=None):
    assert white_list is None or isinstance(white_list, list), type(white_list)

    if white_list is not None:
        white_list = set(item if isinstance(item, str) else item.__qualname__ for item in white_list)

    tree = dict((task.get_qualified_name(), task)
                for task
                in _task_list.values()
                if white_list is None or task.get_qualified_name() in white_list)

    plugins = get_plugin_list()
    for plugin in [plugin for plugin in plugins.values() if white_list is None or plugin.__qualname__ in white_list]:
        tasks = [func
                 for _, func
                 in inspect.getmembers(plugin)
                 if inspect.isfunction(func) and hasattr(func, "yaz_task_config")]
        if len(tasks) == 0:
            continue

        node = tree
        for name in plugin.__qualname__.split("."):
            if not name in node:
                node[name] = {}
            node = node[name]

        for func in tasks:
            node[func.__name__] = Task(plugin_class=plugin, func=func, config=func.yaz_task_config)

    return tree


@decorator
def task(func, **config):
    """Make a method into a Yaz task.

    @task
    def talk(message="Hello World!"):
        return message

    # Or... group multiple tasks together

    # class Tools(Yaz.Plugin):
    #     @task
    #     def say(self, message="Hello World!"):
    #         return message

    #     @task(choices=dict(option=["A", "B", "C"]))
    #     def choose(self, option="A"):
    #         return option
    """
    # @functools.wraps(func)
    # def wrapper(*args, **kwargs):
    #     return func(*args, **kwargs)
        # verbose = self.yaz.verbose
        # if verbose:
        #     start = datetime.datetime.now()
        #     print(self.render("{% color '', '', 'reverse' %}>>> {{ plugin.__class__.__name__ }} {{ task.__name__ }}{% endcolor %} {{ sourcefile }}",
        #                       dict(plugin=self, task=func, sourcefile=inspect.getsourcefile(func))))
        # try:
        #     return func(self, *args, **kwargs)
        # except Exception as exception:
        #     if verbose:
        #         print(self.render("{% color '', '', 'reverse' %}!!! {{ plugin.__class__.__name__ }} {{ exception }}{% endcolor %}",
        #                           dict(plugin=self, task=func, sourcefile=repr(exception))))
        #     raise
        # finally:
        #     if verbose:
        #         stop = datetime.datetime.now()
        #         print(self.render("{% color '', '', 'reverse' %}<<< {{ plugin.__class__.__name__ }} {{ task.__name__ }}{% endcolor %} {{ duration }}",
        #                           dict(plugin=self, task=func, duration=stop - start)))
    if func.__name__ == func.__qualname__:
        assert not func.__qualname__ in _task_list, "Can not define the same task {} twice".format(func.__qualname__)
        _task_list[func.__qualname__] = Task(plugin_class=None, func=func, config=config)
    else:
        func.yaz_task_config = config

    return func



import unittest

from yaz.loader import load


class TestLoader(unittest.TestCase):
    def test_loader(self):
        """Should load a module from a given directory"""
        module = load("./test", "loader")
        self.assertEqual("Done", module.loader_completed)

    def test_unavailable_directory(self):
        """Should not load a module which is not available"""
        module = load("./unavailable_test_directory", "unavailable_test_module_name")
        self.assertIsNone(module)